#ifndef _IPv4_H
#define _IPv4_H

#include <stdint.h>
#include "../eth_base/eth.h"

#define IPv4_ADDR_SIZE 4
#define IPv4_STR_MAX_LENGTH 16
#define IP_ETH_TYPE 0X0800
#define IPv4_MTU 1480 /*Maximo tamaño de payload*/

typedef unsigned char ipv4_addr_t [IPv4_ADDR_SIZE];
typedef struct ipv4_layer ipv4_layer_t;
typedef struct ipv4_packet ipv4_packet_t;

/* Dirección IPv4 a cero "0.0.0.0" */
extern ipv4_addr_t IPv4_ZERO_ADDR;
extern ipv4_addr_t RIP_IP_MULTICAST_ADDR;

/* Logitud máxmima del nombre de un interfaz de red */
#define IFACE_NAME_MAX_LENGTH 32


/* void ipv4_addr_str ( ipv4_addr_t addr, char* str );
 *
 * DESCRIPCIÓN:
 *   Esta función genera una cadena de texto que representa la dirección IPv4
 *   indicada.
 *
 * PARÁMETROS:
 *   'addr': La dirección IP que se quiere representar textualente.
 *    'str': Memoria donde se desea almacenar la cadena de texto generada.
 *           Deben reservarse al menos 'IPv4_STR_MAX_LENGTH' bytes.
 */
void ipv4_addr_str ( ipv4_addr_t addr, char* str );


/* int ipv4_str_addr ( char* str, ipv4_addr_t addr );
 *
 * DESCRIPCIÓN:
 *   Esta función analiza una cadena de texto en busca de una dirección IPv4.
 *
 * PARÁMETROS:
 *    'str': La cadena de texto que se desea procesar.
 *   'addr': Memoria donde se almacena la dirección IPv4 encontrada.
 *
 * VALOR DEVUELTO:
 *   Se devuelve 0 si la cadena de texto representaba una dirección IPv4.
 *
 * ERRORES:
 *   La función devuelve -1 si la cadena de texto no representaba una
 *   dirección IPv4.
 */
int ipv4_str_addr ( char* str, ipv4_addr_t addr );


/*
 * uint16_t ipv4_checksum ( unsigned char * data, int len )
 *
 * DESCRIPCIÓN:
 *   Esta función calcula el checksum IP de los datos especificados.
 *
 * PARÁMETROS:
 *   'data': Puntero a los datos sobre los que se calcula el checksum.
 *    'len': Longitud en bytes de los datos.
 *
 * VALOR DEVUELTO:
 *   El valor del checksum calculado.
 */
uint16_t ipv4_checksum ( unsigned char * data, int len );

/*
 * ipv4_layer_t * ipv4_open(char* ipv4_config_file, char* ipv4_route_table_file);
 *
 * DESCRIPCIÓN:
 *   Esta función inicializa la tabla de rutas, abre el manejador de la
 *   interfaz Ethernet y obtiene nuestra la IPv4 y netmask de la máquina donde origen.
 *
 * PARÁMETROS:
 *    'ipv4_config': Nombre del fichero de configuración
 *    'ipv4_route_table': Nombre del fichero de tabla de rutas
 *
 * ERRORES:
 *   La función devuelve '-1' si se ha producido algún error.
 *
 * VALOR DEVUELTO:
 *   La función devuelve un puntero de tipo ipv4_layer_t con los datos proporcionados
 *	 por un fichero de configuración y una tabla de rutas.
 */

 ipv4_layer_t * ipv4_open(char* ipv4_config_file, char* ipv4_route_table_file);

/*
 *	int ipv4_close(ipv4_layer_t * ipv4_layer )
 *
 * DESCRIPCIÓN:
 *	Esta función cierra la interfaz utilizada por Ethernet,
 *	libera la memoria asignada a la tabla de rutas y
 *	libera la memoria reservada a la que apunta el puntero ipv4_layer.
 *
 * PARÁMETROS:
 *    'ipv4_layer' : Puntero a un espacio de memoria de tamaño del tipo ipv4_layer_t
 *									 donde se guardarán temporalmente los datos para trabajar con dicha estructura
 *									 Posteriormente será liberada en la función ipv4_close
 *
 * VALOR DEVUELTO:
 *   La función devuelve 0 si se ha cerrado correctamente, ó -1 en caso de error.
 *
 * ERRORES:
 *   La función devuelve '-1' si se ha producido algún error.
 */

 int ipv4_close ( ipv4_layer_t * ipv4_layer );

 /*
  *	int ipv4_send
  *	( ipv4_layer_t * ipv4, ipv4_addr_t ipv4_addr_dest,
  *	uint8_t protocol, unsigned char *data, int data_length)
  *
  * DESCRIPCIÓN:
  *   Esta función envía la información de la capa superior a la capa inferior (Ethernet)
  *
  * PARÁMETROS:
  *		'ipv4_layer'			: Estructura ipv4_layer_t con los datos de la máquina que envía el paquete IP
  *    'ipv4_addr_dest'	: IP destino
  *    'protocol'				: Tipo de protocolo.
  *    'data'						: Datos a enviar.
  *    'length'					: Longitud de los datos que vamos a enviar.
  *
  * VALOR DEVUELTO:
  *   El número de bytes de datos que han podido ser enviados.
  *
  * ERRORES:
  *   La función devuelve '-1' si se ha producido algún error.
  */

	int ipv4_send( ipv4_layer_t * ipv4_layer, ipv4_addr_t ipv4_addr_dest, uint8_t protocol, unsigned char *data, int data_length);

/* int ipv4_receive
 * 	(	ipv4_layer_t * ipv4, ipv4_addr_t ipv4_addr_src,
 *		uint8_t protocol, unsigned char buffer[], int buf_len, long int timeout	);
 *
 * DESCRIPCIÓN:
 *   Esta función recibe la información de la capa superior
 *
 * PARÁMETROS:
 *		'ipv4_layer'		: Estructura ipv4_layer_t con los datos de la máquina que recibe el paquete IP
 *    'ipv4_addr_src'	: IP origen del servidor que envió el paquete IP
 *    'protocol'			: Tipo de protocolo.
 *    'buffer'				: Array donde se almacenarán los datos de la paquete recibido.
 *    'buf_len'				: Longitud del 'buffer' dónde se almacenarán los datos del paquete
 *             					recibido. Si se reciben más datos de los que caben el en 'buffer'
 *             					sólo se copiaran los primeros 'buf_len' bytes.
 *    'timeout' 			: Tiempo en milisegundos que debe esperarse a recibir un paquete
 *             					antes de retornar. Un número negativo indicará que debe
 *             					esperarse indefinidamente, mientras que con un '0' la función
 *             					retornará inmediatamente, se haya recibido o no una trama.
 *
 * VALOR DEVUELTO:
 *   La longitud en bytes de los datos del paquete recibido o '0' si no se ha recibido ninguna
 *   trama porque ha expirado el temporizador.
 *
 * ERRORES:
 *   La función devuelve '-1' si se ha producido algún error.
 */

int ipv4_receive( ipv4_layer_t * ipv4_layer, ipv4_addr_t ipv4_addr_src, uint8_t protocol, unsigned char buffer[], int buf_len, long int timeout);

/* void ipv4_getaddr ( ipv4_layer_t * ipv4_layer, ipv4_addr_t addr );
 *
 * DESCRIPCIÓN:
 *   Esta función permite obtener la dirección IP de la máquina especificada.
 *
 * PARÁMETROS:
 *   'ipv4_layer': Estructura ipv4_layer_t con los datos de la máquina
 *            		La estructura debe haber sido inicializada con 'ipv4_open()'
 *            		previamente.
 *    'addr': Array donde se copiará la dirección IP de la estructura ipv4_layer.
 *
 */

void ipv4_getaddr ( ipv4_layer_t * ipv4_layer, ipv4_addr_t addr );

/* char * ipv4_getiface_name (ipv4_layer_t * ipv4_layer);
 *
 * DESCRIPCIÓN:
 *   Esta función permite obtener el nombre de la iface de una estructura ipv4_layer_t especificada.
 *
 * PARÁMETROS:
 *   'ipv4_layer': Estructura ipv4_layer_t con los datos de la máquina
 *            		La estructura debe haber sido inicializada con 'ipv4_open()'
 *            		previamente.
 * VALOR DEVUELTO:
 *   Cadena de texto con el nombre de la interfaz.
 *
 */

char * ipv4_getiface_name (ipv4_layer_t * ipv4_layer);

#endif /* _IPv4_H */
