#ifndef _RIP_H
#define _RIP_H

#include "../ipv4_base/ipv4.h"
#include "../udp/udp.h"

#define RIP_PORT_UDP 520 /* Puerto de RIP sobre UDP (servidor). */
#define RIP_COMMAND_REQUEST 0x01
#define RIP_COMMAND_RESPONSE 0x02
#define RIP_VERSION 0x02
#define RIP_INF_METRIC 16

#define RIP_HEADER_LENGTH 4
#define RIP_MAX_LENGTH 504 /* 4 + 20*25 */
#define RIP_ENTRY_LENGTH 20
#define RIP_MAX_ENTRIES 25


/*
    Estructura de un paquete RIPv2, según la RFC2453: https://tools.ietf.org/html/rfc2453

	0                   1                   2                   3
	0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	|  command (1)  |  version (1)  |       must be zero (2)        |
	+---------------+---------------+-------------------------------+
	|                                                               |
	~                         RIP Entry (20)                        ~
	|                                                               |
	+---------------+---------------+---------------+---------------+

	The same header format is used for RIP-1 and RIP-2 messages (see section 3.4).
	The format for the 20-octet route entry (RTE) for RIP-2 is:

 	0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   | Address Family Identifier (2) |        Route Tag (2)          |
   +-------------------------------+-------------------------------+
   |                         IP Address (4)                        |
   +---------------------------------------------------------------+
   |                         Subnet Mask (4)                       |
   +---------------------------------------------------------------+
   |                         Next Hop (4)                          |
   +---------------------------------------------------------------+
   |                         Metric (4)                            |
   +---------------------------------------------------------------+

	1 - request
					A request for the responding system to send all or
                  	part of its routing table.

	2 - response
					A message containing all or part of the sender's
					routing table.  This message may be sent in response
					to a request, or it may be an unsolicited routing
					update generated by the sender.

	The Request is processed entry by entry.  If there are no entries, no
	response is given.  There is one special case.  If there is exactly
	one entry in the request, and it has an address family identifier of
	zero and a metric of infinity (i.e., 16), then this is a request to
	send the entire routing table.
*/

struct rip_entry {
	uint16_t addr_family_id; /* El Address Family ID sobre IPv4 es 2 siempre */
	uint16_t route_tag; /* Siempre a 0 para diferenciar rutas aprendidas por RIP o por protocolos externos */
	ipv4_addr_t ipv4_addr;
	ipv4_addr_t subnet_mask;
	ipv4_addr_t next_hop; /* NOTA! El siguiente salto soy yo (0.0.0.0) siempre que anuncio rutas aprendidas de routers RIP,
							 en caso de que el siguiente salto no hable RIP será él el siguiente salto. */
	uint32_t metric;
};

typedef struct rip_entry rip_entry_t;


struct rip_packet {
    uint8_t command; /* Request o response */
    uint8_t version; /* RIPv2 */
    uint16_t zeros; /* 0's */
    struct rip_entry entries [RIP_MAX_ENTRIES]; /* Array de entradas */
};

typedef struct rip_packet rip_packet_t;


int rip_len2entries(int len);
int rip_entries2len(int entries);
void rip_print_packet(rip_packet_t * rip_packet, int length);

#endif /* _RIP_H */
