#include "udp.h"

#include "../ipv4_config/ipv4_config.h"
#include "../ipv4_route_table/ipv4_route_table.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <timerms.h>
#include <time.h>
#include <libgen.h>

#define UDP_PROTO_IP 17 /* Tipo de protocolo UDP sobre IPv4. */
#define UDP_PSEUDOHEADER_LENGTH 12 /* Tamaño de la pseudocabecera UDP (checksum). */
#define UDP_CHECKSUM_ZEROS 0x00 /* Zeros para el checksum */


/*

      0      7 8     15 16    23 24    31
      +--------+--------+--------+--------+
      |     Source      |   Destination   |
      |      Port       |      Port       |
      +--------+--------+--------+--------+
      |                 |                 |
      |     Length      |    Checksum     |
      +--------+--------+--------+--------+
      |
      |          data octets ...
      +---------------- ...

          User Datagram Header Format

*/

struct udp_datagram {
  uint16_t src_port_udp; /* Puerto origen */
  uint16_t dst_port_udp; /* Puerto destino */
  uint16_t udp_length; /* Longitud total del datagrama UDP */
  uint16_t checksum; /* El campo checksum es una suma de comprobación de 16 bits que abarca
                        una pseudo-cabecera IP (con las IP origen y destino, el protocolo y
                        la longitud del paquete UDP), la cabecera UDP, los datos y 0's hasta
                        completar un múltiplo de 16.

                        Checksum is the 16-bit one's complement of the one's complement sum of a
                        pseudo header of information from the IP header, the UDP header, and the
                        data,  padded  with zero octets  at the end (if  necessary)  to  make  a
                        multiple of two octets.*/
  unsigned char payload [MTU_UDP];
};

/*
              pseudo-cabecera IP

      0      7 8     15 16    23 24    31
      +--------+--------+--------+--------+
      |          source address           |
      +--------+--------+--------+--------+
      |        destination address        |
      +--------+--------+--------+--------+
      |  zero  |protocol|   UDP length    |
      +--------+--------+--------+--------+

*/

/*
Estructura que combina los campos de una ip pseudoheader de UDP y un miembro de tipo udp_datagram_t.
Esto nos ayuda a crear un manejador sencillo para realizar el checksum de UDP.
phd (pseudoheader + datagram) ;)
*/

struct phd_layer {
  ipv4_addr_t ipv4_addr_src; /* Dirección IP origen */
  ipv4_addr_t ipv4_addr_dst; /* Dirección IP destino */
  uint8_t zero; /* Zeros */
  uint8_t protocol; /* UDP Protocol */
  uint16_t udp_length; /* UDP length */
  udp_datagram_t datagram;
};

struct udp_layer {
  ipv4_layer_t * ipv4_layer;
  uint16_t port;
};

/*
 * udp_layer_t udp_open(char* ipv4_config_file, char* ipv4_route_table_file, uint16_t port)
 *
 * DESCRIPCIÓN:
 *   Esta función crea, durante el tiempo de comunicación, un manejador de la capa UDP el cual contiene
 *   los datos de un manejador de la capa IP y un puerto.
 *
 * PARÁMETROS:
 *    'ipv4_config_file': Nombre del fichero de onfiguración
 *    'ipv4_route_table_file': Nombre del fichero de tabla de rutas
 *    'port': Puerto propio de UDP.
 *
 * ERRORES:
 *   La función devuelve 'NULL' si se ha producido algún error.
 *
 * VALOR DEVUELTO:
 *   La función devuelve un puntero de tipo udp_layer_t
 */

udp_layer_t * udp_open(char* ipv4_config_file, char* ipv4_route_table_file, uint16_t port)
{
  if(port == 0){
    unsigned int seed = time(NULL);
    srand(seed);
    float random = UDP_MAX_PORT - UDP_MIN_PORT;
    /* Generar número aleatorio */
    port = UDP_MIN_PORT+(int) (random*rand()/(RAND_MAX + random));
  }
  udp_layer_t * udp_layer = malloc(sizeof(udp_layer_t));

  udp_layer->ipv4_layer = ipv4_open(ipv4_config_file, ipv4_route_table_file);
  udp_layer->port = port;

  return udp_layer;
}

/*
 * int udp_close(udp_layer_t * udp_layer)
 *
 * DESCRIPCIÓN:
 *   Esta función libera la memoria del manejador de la capa IP llamando a la funcion "ipv4_close"
 *   y libera la memoria reservada para el manejador UDP creado en udp_open.
 *
 * PARÁMETROS:
 *    'udp_layer' : Puntero a un espacio de memoria de tamaño del tipo udp_layer_t
 *									 donde se guardara temporalmente los datos para trabajar con dicha estructura
 *									 Dicho espacio será liberado en esta función.
 *
 * VALOR DEVUELTO:
 *   La función devuelve -1 si ha habido algún error, 0 si se ha cerrado correctamente.
 *
 * ERRORES:
 *   La función devuelve '-1' si se ha producido algún error.
 */

int udp_close(udp_layer_t * udp_layer)
{
  int err = ipv4_close(udp_layer->ipv4_layer); /* Cerrar IP y capas inferiores */
  if(err == -1){
    printf("Error en udp_close.\n");
    return -1;
  }
  free(udp_layer);
  return 0;
}

/*
 * int udp_send(udp_layer_t * udp_layer, ipv4_addr_t ip_dst, uint16_t port_dst, unsigned char* data, int length)
 *
 * DESCRIPCIÓN:
 *   Esta función envía la información de la capa superior a la capa inferior (IP)
 *
 * PARÁMETROS:
 *    'udp_layer': manejador de la capa UDP
 *    'ip_dst': IP destino donde se quiere enviar el datagrama
 *    'port_dst': Puerto UDP de destino donde se quiere enviar el datagrama.
 *    'data': Datos a enviar.
 *    'length': Longitud de los datos que vamos a enviar.
 *
 * VALOR DEVUELTO:
 *   El número de bytes de datos que han podido ser enviados.
 *
 * ERRORES:
 *   La función devuelve '-1' si se ha producido algún error.
 */

int udp_send(udp_layer_t * udp_layer, ipv4_addr_t ip_dst, uint16_t port_dst, unsigned char* data, int length)
{
  int total_length = length + UDP_HEADER_LENGTH; /* Logintud total del datagrama, incluida cabecera */
  int checksum_length = total_length + UDP_PSEUDOHEADER_LENGTH; /* Longitud para hacer el checksum */

  ipv4_addr_t addr;
  ipv4_getaddr (udp_layer->ipv4_layer, addr);

  phd_layer_t pseudoheader;

  pseudoheader.datagram.src_port_udp = htons(udp_layer->port);
  pseudoheader.datagram.dst_port_udp = htons(port_dst);
  pseudoheader.datagram.udp_length = htons(total_length);
  pseudoheader.datagram.checksum = htons(0x0000);
  memset(pseudoheader.datagram.payload, 0x00, MTU_UDP);
  memcpy(pseudoheader.datagram.payload, data, length);

  /*Rellenamos la pseudo cabecera para crear el checksum completo */
  memcpy(pseudoheader.ipv4_addr_src, addr, IPv4_ADDR_SIZE);
  memcpy(pseudoheader.ipv4_addr_dst, ip_dst, IPv4_ADDR_SIZE);
  pseudoheader.zero = 0x00;
  pseudoheader.protocol = UDP_PROTO_IP;
  pseudoheader.udp_length = htons(total_length);

  pseudoheader.datagram.checksum = htons(ipv4_checksum((unsigned char *)&pseudoheader, checksum_length)); /* Actualizando el valor del checksum */

  int sent = ipv4_send(udp_layer->ipv4_layer, ip_dst, UDP_PROTO_IP, (unsigned char *) &pseudoheader.datagram, total_length);

  if(sent != total_length){
		printf("Error al enviar UDP desde IPv4.\n");
		return -1;
	}

  return (sent - UDP_HEADER_LENGTH);
}

/*
 * int udp_receive(udp_layer_t * udp_layer, uint16_t port_src, unsigned char buffer[], int buf_len, long int timeout)
 *
 * DESCRIPCIÓN:
 *   Esta función recibe la información de la capa superior
 *
 * PARÁMETROS:
 *    'udp_layer'     : manejador de la capa UDP
 *    'ipv4_addr_src' : IP origen del envio del datagrama UDP
 *    'port_src'      : Puerto UDP origen del datagrama recibido (parámetro de salida).
 *    'buffer'        : Array donde se almacenarán los datos de la trama recibida.
 *    'buf_len'       : Longitud del 'buffer' dónde se almacenarán los datos de la trama
 *                      recibida. Si se reciben más datos de los que caben el en 'buffer'
 *                      sólo se copiaran los primeros 'buf_len' bytes.
 *    'timeout'       : Tiempo en milisegundos que debe esperarse a recibir una trama
 *                      antes de retornar. Un número negativo indicará que debe
 *                      esperarse indefinidamente, mientras que con un '0' la función
 *                      retornará inmediatamente, se haya recibido o no una trama.
 *
 * VALOR DEVUELTO:
 *   La longitud en bytes de los datos de la trama recibida
 *
 * ERRORES:
 *   La función devuelve '-1' si se ha producido algún error.
 */

int udp_receive(udp_layer_t * udp_layer, ipv4_addr_t ipv4_addr_src, uint16_t * port_src, unsigned char buffer[], int buf_len, long int timeout)
{
  int buf_total_len = buf_len + UDP_HEADER_LENGTH; /* Cantidad total de datos a recibir de IPv4 */
  int received = 0;
  udp_datagram_t * datagram_ptr;
  unsigned char buffer_recv[buf_total_len];
  timerms_t timer;
	timerms_reset(&timer, timeout);
  int correct_port = 0;
  ipv4_addr_t addr;
  ipv4_getaddr (udp_layer->ipv4_layer, addr);

  do {
    long int time_left = timerms_left(&timer);
    received = ipv4_receive(udp_layer->ipv4_layer, ipv4_addr_src, UDP_PROTO_IP, buffer_recv, buf_total_len, time_left);

    /* Comprobación de errores */
	  if (received == -1) {
	  	 printf("Error en UDP al ipv4_receive.\n");
       return -1;
	  } else if (received == 0) {
      printf(" UDP Timeout.\n");
			return 0;
		} else if (received < UDP_HEADER_LENGTH) {
      printf( "UDP Datagrama de tamaño invalido.\n");
      continue;
    }

    datagram_ptr = (udp_datagram_t *)buffer_recv;

    //uint16_t checksum_recv = ntohs(datagram_ptr->checksum);
    int recv_length = ntohs(datagram_ptr->udp_length);
    datagram_ptr->checksum = htons(0x0000);

    phd_layer_t pseudoheader;

    /* Checksum pseudoheader */
    memcpy(pseudoheader.ipv4_addr_src, ipv4_addr_src, IPv4_ADDR_SIZE);
    memcpy(pseudoheader.ipv4_addr_dst, addr, IPv4_ADDR_SIZE);
    pseudoheader.zero = 0x00;
    pseudoheader.protocol = UDP_PROTO_IP;
    pseudoheader.udp_length = htons(recv_length);

    /* Checksum datagram */
    memset(pseudoheader.datagram.payload, 0x00, MTU_UDP);
    memcpy(&pseudoheader.datagram, datagram_ptr, recv_length);

    datagram_ptr->checksum = ipv4_checksum((unsigned char *)&pseudoheader, recv_length + UDP_PSEUDOHEADER_LENGTH);

/*
    if (datagram_ptr->checksum != checksum_recv) {
      printf( "UDP Checksum inválido.\n");
      continue;
    }
*/
    correct_port = (ntohs(datagram_ptr->dst_port_udp) == udp_layer->port);

  } while (!correct_port);

  * port_src = ntohs(datagram_ptr->src_port_udp);

  int payload_len = ntohs(datagram_ptr->udp_length) - UDP_HEADER_LENGTH;

  if (buf_len > payload_len) {
      buf_len = payload_len;
  }

  memcpy(buffer, datagram_ptr->payload, payload_len);

  return payload_len;
}

/* char * udp_getiface_name (udp_layer_t * udp_layer);
 *
 * DESCRIPCIÓN:
 *   Esta función permite obtener el nombre de la iface de una estructura udp_layer_t especificada.
 *
 * PARÁMETROS:
 *   'udp_layer': Estructura udp_layer_t con los datos de la máquina
 *            		La estructura debe haber sido inicializada con 'ipv4_open()'
 *            		previamente.
 * VALOR DEVUELTO:
 *   Cadena de texto con el nombre de la interfaz.
 *
 */

char * udp_getiface_name (udp_layer_t * udp_layer)
{
	char* iface_name = NULL;

	if (udp_layer != NULL) {
		iface_name = ipv4_getiface_name(udp_layer->ipv4_layer);
  }
	return iface_name;
}
